from ast import Call
import asyncio
import serial_asyncio
import time
from enum import IntEnum
from typing import Any, Callable, List, Tuple
import logging
import serial

from .types import SerialOrNetwork, EnergyReading
from .comm import HanazederPacket, HanazederReader, hanazeder_encode_msg, hanazeder_decode_num
from .encoding import dec_to_bytes, byte_to_hex

logger = logging.getLogger('hanazeder')

class ConnectError(Exception):
    pass

class NotConnectedError(Exception):
    pass

class ConnectionInvalidError(Exception):
    pass

class RequestTimeoutError(Exception):
    pass

class ShutdownError(Exception):
    pass
class DeviceType(IntEnum):
    FP10 = 0
    FP6 = 1
    FP3 = 2
    FP2 = 3
    FP1 = 4
    SH3 = 5
    SH2 = 6
    SH1 = 7

class HardwarePlatform(IntEnum):
    FP10 = 0
    FP3 = 1

SENSOR_LABELS = [
    None,
    "Kollektor",
    "Sonnenf.",
    "Koll-RL",
    "Boiler",
    "Boiler 1",
    "Boiler 2",
    "Boiler/U",
    "Boil.1/U",
    "Boil.2/U",
    "Boiler/M",
    "Boil.1/M",
    "Boil.2/M",
    "Boiler/O",
    "Boil.1/O",
    "Boil.2/O",
    "Puffer",
    "Puffer 1",
    "Puffer 2",
    "Puffer/U",
    "Puff.1/U",
    "Puff.2/U",
    "Puffer/M",
    "Puff.1/M",
    "Puff.2/M",
    "Puffer/O",
    "Puff.1/O",
    "Puff.2/O",
    "Speicher",
    "Speich.1",
    "Speich.2",
    "Speich/U",
    "Spei.1/U",
    "Spei.2/U",
    "Speich/M",
    "Spei.1/M",
    "Spei.2/M",
    "Speich/O",
    "Spei.1/O",
    "Spei.2/O",
    "Unten",
    "Mitte",
    "Oben",
    "Vorlauf",
    "Ruecklauf",
    "Warmwass.",
    "Kaltwass.",
    "Tauscher",
    "Plattent.",
    "Heizkreis",
    "Heizkrs.1",
    "Heizkrs.2",
    "Heizkrs.3",
    "Heizkoerp",
    "Bodenheiz",
    "Heiz-Kes.",
    "Kessel 1",
    "Kessel 2",
    "Holz-Kes.",
    "Oel-Kess.",
    "Gas-Kess.",
    "Waermepu.",
    "Pelletsk.",
    "Hackgutk.",
    "Brenner",
    "Ofen",
    "Schwimmb.",
    "Becken",
    "Bad",
    "Raumtemp.",
    "Raumtmp.1",
    "Raumtmp.2",
    "Aussentmp",
    "Keller",
    "Erdgesch.",
    "1.Stock",
    "2.Stock",
    "3.Stock",
    "Koll-Sued",
    "Koll-West",
    "Koll-Ost",
    "Raum-Regl.",
    "VL Solar",
    "RL Solar",
    "RL Kessel",
    "VL Heizk.",
    "VL Bodenh",
    "VL Wandh.",
    "Wintergar",
    "Treibhaus",
    "Nicht bel"
]

DecoderCB = Callable[[HanazederPacket], Any]
class ConfigEntry:
    key = 0

    def __init__(self, key, data: bytes):
        self.key = key
        self.value = data[0]
        self.max_value = data[1]
        self.min_value = data[2]
        self.change_step = data[3]

class HanazederRequestState(IntEnum):
    UNSENT = 0
    SENT = 1
    SUCCESSFUL = 2
    TIMEOUT = 3
    SHUTDOWN = 4
    DISCONNECTED = 5
class HanazederRequest:
    result = None
    def __init__(self, msg_no: int, type: int, decoder: DecoderCB, msg: bytes):
        self.msg_no = msg_no
        self.type = type
        self.decoder = decoder
        self.created = time.monotonic()
        self.event = asyncio.Event()
        self.msg = msg
        self.state = HanazederRequestState.UNSENT

class FPProtocol(asyncio.Protocol):
    device = None
    def connection_made(self, transport):
        self.connection = transport
        if hasattr(self.connection, 'serial'):
            transport.serial.rts = True

    def data_received(self, data):
        if self.device.debug:
            logger.debug('Data received: %s', data)
        self.device.read_bytes(data)
        logger.debug('Done with read_bytes')
        if hasattr(self.connection, 'serial'):
            self.connection.resume_reading()
            self.connection.serial.rts = True
        #asyncio.get_event_loop().create_task(self.device.read_bytes(data))

    def connection_lost(self, exc):
        logger.warn('The server closed the connection')
        self.device.connected = False
        # Awake all listeners
        for request in self.device.queue:
            request.state = HanazederRequestState.DISCONNECTED
            request.event.set()

    def pause_writing(self):
        print('pause writing')
        print(self.transport.get_write_buffer_size())

    def resume_writing(self):
        print(self.transport.get_write_buffer_size())
        print('resume writing')

class HanazederFP:
    HEADER = b'\xEE'
    last_msg_num = 0
    connected = True
    debug = False
    connection: SerialOrNetwork
    queue: List[HanazederRequest] = []
    running = True

    def __init__(self, debug=False, request_timeout=2):
        self.debug = debug
        self.loop = asyncio.get_running_loop()
        self.queue_lock = asyncio.Lock()
        self.msg_no_lock = asyncio.Lock()
        self.request_timeout = request_timeout
    
    async def open(self,
            serial_port="/dev/ttyUSB0",
            address=None,
            port=None,
            timeout=1000):
        if serial_port:
            (self.connection, proto) = await serial_asyncio.create_serial_connection(
                self.loop, FPProtocol, serial_port, baudrate=38400, timeout = timeout,
                rtscts = True, bytesize=serial.EIGHTBITS, stopbits = serial.STOPBITS_ONE, parity=serial.PARITY_NONE 
            )
        elif address and port:
            (self.connection, proto) = await self.loop.create_connection(FPProtocol, address, port)
        else:
            raise ConnectionInvalidError("Specify either address and port or serial port")
        proto.device = self
        self.reader = HanazederReader(self.connection, self.HEADER, self.debug)
        # Check queue for stuck messages periodically
        self.queue_check_task = self.loop.create_task(self.check_queue(), name="check_queue")
    
    
    async def get_next_msg_no(self) -> int:
        async with self.msg_no_lock:
            msg_no = self.last_msg_num
            self.last_msg_num = (self.last_msg_num + 1) % 256
            # Avoid escape byte
            if self.last_msg_num == 238:
                self.last_msg_num = 239
            return msg_no
    
    async def send_msg(self, msg: bytes, decoder: DecoderCB) -> HanazederRequest:
        if self.debug:
            logger.debug('Sending msg #%d: %s', msg[1], byte_to_hex(msg))
        request = HanazederRequest(msg[1], msg[2], decoder, msg)
        async with self.queue_lock:
            self.queue.append(request)
            self.connection.write(msg)
            if hasattr(self.connection, 'serial'):
                self.connection.serial.flush()
                logger.debug('flushing serial')
            request.state = HanazederRequestState.SENT
            return request
    
    def read_bytes(self, bytes):
        logger.debug('read_bytes')
        for byte in bytes:
            packet = self.reader.read(byte)
            if packet:
                self.handle_packet(packet)
    
    async def check_queue(self):
        while self.connected:
            now = time.monotonic()
            async with self.queue_lock:
                logger.debug('queue check starting, getting lock')
                for index, request in enumerate(self.queue):
                    if now - request.created > self.request_timeout:
                        logger.warn('Request #%d has timed out', request.msg_no)
                        #request.msg_no = await self.get_next_msg_no()
                        #request.msg = hanazeder_encode_msg(self.HEADER, request.msg_no, request.msg[2:-1])
                        self.connection.write(request.msg)
                        if hasattr(self.connection, 'serial'):
                            self.connection.serial.flush()
                            logger.debug('flushing serial')
                        if self.debug:
                            logger.debug('Resending msg #%d: %s', request.msg[1], byte_to_hex(request.msg))
                        #request.state = HanazederRequestState.TIMEOUT
                        #request.event.set()
                        #self.queue.pop(index)
            logger.debug('queue check done')
            await asyncio.sleep(self.request_timeout)
          

    def shutdown(self):
        self.running = False       
        for request in self.queue:
            request.state = HanazederRequestState.SHUTDOWN
            request.event.set()


    def handle_packet(self, packet: HanazederPacket):
        if self.debug:
            packet_debug = ""
            for req in self.queue:
                packet_debug = f"{packet_debug} #{req.msg_no}"
            logger.debug('Packet read #%d type %d: %s.', packet.msg_no, packet.msg_type, packet.msg)
            logger.debug('Queue: %s', packet_debug)
        found = False
        #async with self.queue_lock:
        for index, req in enumerate(self.queue):
            if req.msg_no == packet.msg_no:
                req.result = req.decoder(packet)
                found = True
                req.state = HanazederRequestState.SUCCESSFUL
                req.event.set()
                self.queue.pop(index)
                break
        if not found:
            logger.error(f"Couldn't find message {packet.msg_no} in queue!")

    async def handle_req_response(self, req: HanazederRequest):
        if req.state == HanazederRequestState.TIMEOUT:
            raise RequestTimeoutError()
        elif req.state == HanazederRequestState.SHUTDOWN:
            raise ShutdownError()
        elif req.state == HanazederRequestState.DISCONNECTED:
            raise NotConnectedError()
    
    async def create_read_information_msg(self) -> bool:
        return hanazeder_encode_msg(self.HEADER, await self.get_next_msg_no(), b'\x01\x00')
    
    async def read_information(self):
        req = await self.send_msg(await self.create_read_information_msg(), self.parse_information_packet)
        await req.event.wait()
        await self.handle_req_response(req)
    
    def parse_information_packet(self, msg: HanazederPacket):
        response = msg.msg
        # 00 f0 05 00 00 02 01 06 f9
        self.device_type = DeviceType(response[0])
        self.hardware_platform = HardwarePlatform(response[1])
        self.connection_flags = response[2]
        if (len(response) >= 5):
            self.version = f'{response[3]}.{response[4]}'
    
    async def create_read_sensor_msg(self, idx: int) -> bytes:
        request_bytes = bytes(b'\x04\x01') + idx.to_bytes(1, byteorder='little')
        return hanazeder_encode_msg(self.HEADER, await self.get_next_msg_no(), request_bytes)
    
    async def read_sensor(self, idx: int) -> float:
        if not self.connected:
            raise NotConnectedError()
        
        req = await self.send_msg(await self.create_read_sensor_msg(idx), self.parse_sensor_packet)
        await req.event.wait()
        await self.handle_req_response(req)
        return req.result
    
    def parse_sensor_packet(self, msg: HanazederPacket) -> float:
        value = hanazeder_decode_num(self.HEADER, msg.msg)
        return value
    
    async def create_read_config_block_msg(self, start: int, count: int) -> bytes:
        request = bytes(b'\x07\x03') + start.to_bytes(2, byteorder='little') + count.to_bytes(1, byteorder='little')
        return hanazeder_encode_msg(self.HEADER, await self.get_next_msg_no(), request)
    
    async def read_config_block(self, start: int, count: int) -> List[ConfigEntry]:
        if not self.connected:
            raise NotConnectedError()
        
        def parse_config_block_packet(msg: HanazederPacket) -> List[ConfigEntry]:
            response = msg.msg
            entries = []
            for x in range(0, int(round(len(response) / 4))):
                chunk = response[x * 4: (x + 1) * 4]
                entry = ConfigEntry(start + x, chunk)
                entries.append(entry)
            return entries
        req = await self.send_msg(await self.create_read_config_block_msg(start, count), parse_config_block_packet)
        await req.event.wait()
        await self.handle_req_response(req)
        return req.result

    async def create_read_sensor_name_msg(self, idx: int) -> bytes:
        request = bytes(b'\x13\x01') + idx.to_bytes(1, byteorder='little')
        return hanazeder_encode_msg(self.HEADER, await self.get_next_msg_no(), request)
    
    async def read_sensor_name(self, idx: int) -> str:
        if not self.connected:
            raise NotConnectedError()

        req = await self.send_msg(await self.create_read_sensor_name_msg(idx), self.parse_sensor_name_packet)
        await req.event.wait()
        await self.handle_req_response(req)
        return req.result
    
    def parse_sensor_name_packet (self, msg: HanazederPacket) -> str:
        if msg.msg and len(msg.msg) > 1:
            return msg.msg[1:].decode('ascii', errors='ignore').strip()
    
    async def create_read_debug_block_msg(self, start: int, count: int) -> bytes:
        request = bytes(b'\x20\x03') + dec_to_bytes(start) + count.to_bytes(1, byteorder='little')
        return hanazeder_encode_msg(self.HEADER, await self.get_next_msg_no(), request)
    
    async def read_debug_block(self, start: int, count: int, decoder: DecoderCB) -> Any:
        if not self.connected:
            raise NotConnectedError()
        req = await self.send_msg(await self.create_read_debug_block_msg(start, count), decoder)
        await req.event.wait()
        await self.handle_req_response(req)
        return req.result
    
    async def read_energy(self) -> EnergyReading:
        return await self.read_debug_block(313, 8, self.parse_energy_packet)
    
    def parse_energy_packet(self, msg: HanazederPacket) -> Tuple[int, int, int]:
        total = hanazeder_decode_num(self.HEADER, msg.msg[0:2])
        current = hanazeder_decode_num(self.HEADER, msg.msg[2:4])
        impulse = hanazeder_decode_num(self.HEADER, msg.msg[4:6])
        return (total, current, impulse)
    




import asyncio
from typing import List
from .Hanazeder import ConfigEntry, HanazederFP, SENSOR_LABELS
import argparse
import sys
import logging

FORMAT = '%(asctime)s %(message)s'
logging.basicConfig(format=FORMAT, level=logging.DEBUG)

class CliReader:
    sensor_vals = [None] * 15
    sensor_names = [None] * 15
    sensor_custom_name = None
    sensor_idx = None

    def config_block_read(self, configs: List[ConfigEntry]):
        for i, config_label in enumerate(configs):
            if config_label.value > 0:
                self.sensor_names[i] = SENSOR_LABELS[config_label.value]

    def print_sensor(self, i):
        print (f'Sensor {self.sensor_names[i]} ({i}) has value {self.sensor_vals[i]}')
    
    async def get_sensor_name(self, i):
        return self.sensor_names[i]

    async def main(self) -> int:
        parser = argparse.ArgumentParser()
        parser.add_argument("--serial-port", help="set serial port",
                        type=str)
        parser.add_argument("--sensor", help="sensor to read",
                        type=int)
        parser.add_argument("--sensors", help="read all sensors", action="store_true")
        parser.add_argument("--energy", help="read energy values", action="store_true")
        parser.add_argument("--loop", help="read values in a loop", action="store_true")
        parser.add_argument("--debug", help="print low-level messages", action="store_true")
        parser.add_argument("--address", help="connect to HOSTNAME, needs port as well",
                        type=str)
        parser.add_argument("--port", help="connect to HOSTNAME on port PORT",
                        type=int, default=5000)
        args = parser.parse_args()

        if args.address and args.serial_port:
            print('Cannot specify both serial-port and address')
            return 1
        
        if args.address and not args.port:
            print('Specify port together with address')
            return 2
        if not args.sensor and not args.energy and not args.sensors:
            print("Don't know what to do, please add --energy, --sensors and/or --sensor")

        self.conn = HanazederFP(debug=args.debug, request_timeout=2)
        await self.conn.open(serial_port=args.serial_port, address=args.address, port=args.port, timeout=0.2)
        await self.conn.read_information()
        print(f'Connected to {self.conn.device_type.name} with version {self.conn.version}')

        loop_count = 1000 if args.loop else 1
        
        async def dummy():
            return None
        
        while loop_count > 0:
            loop_count = loop_count - 1
            if args.energy:
                energy = await self.conn.read_energy()
                print('Energy readings:')
                print(f'  Total   {energy[0]}')
                print(f'  Current {energy[1]}')
                print(f'  Impulse {energy[2]}')

            if args.sensors:
                # Read label from fixed list
                name_tasks = []
                
                self.config_block_read(await self.conn.read_config_block(27, 15))
                for i in range(0, 15):
                    # Also read custom name
                    if self.sensor_names[i] is None:
                        self.sensor_names[i] = await self.conn.read_sensor_name(i)
                    #else:
                    #    values_tasks.append(self.get_sensor_name(i))
                values_tasks = []
                for i in range(0, 15):
                    values_tasks.append(self.conn.read_sensor(i))
                
                self.sensor_vals = await asyncio.gather(*values_tasks)
                
                for i in range(0, 15):
                    self.print_sensor(i)
            elif args.sensor is not None:
                self.sensor_idx = args.sensor - 1

                print(f'Reading sensor {args.sensor}')
                self.sensor_vals[self.sensor_idx] = await self.conn.read_sensor(self.sensor_idx)
                # Read label from fixed list
                self.config_block_read(await self.conn.read_config_block(27, 15))
                # Also read custom name
                if self.sensor_names[self.sensor_idx] is None:
                    self.sensor_names[self.sensor_idx] = await self.conn.read_sensor_name(self.sensor_idx)
                self.print_sensor(self.sensor_idx)
                
        return 0

if __name__ == '__main__':
    instance = CliReader()
    sys.exit(asyncio.run(instance.main(), debug=True)
)



from enum import IntEnum
from .types import SerialOrNetwork
from .encoding import *

import logging
from crccheck.crc import Crc8Maxim

logger = logging.getLogger('hanazeder.comm')

SENSOR_GONE = b'\xFF\x7F'

class ReadtimeoutException(Exception):
    pass

class ChecksumNotMatchingException(Exception):
    pass

class InvalidHeaderException(Exception):
    pass

class IllegalArgumentException(Exception):
    pass

class ReaderState(IntEnum):
    LOOKING_FOR_HEADER = 0
    EXPECTING_MSG_NO = 1
    EXPECTING_TYPE = 2
    EXPECTING_AFTER_TYPE = 3
    EXPECTING_SIZE = 4
    READING_PAYLOAD = 5
    EXPECTING_CHECKSUM = 6
    ESCAPING = 7


class HanazederPacket:
    msg_no = None
    msg_type = None
    msg_size = None
    msg = None
class HanazederReader:
    state = ReaderState.LOOKING_FOR_HEADER
    state_before_escaping = None
    packet: HanazederPacket = None

    def __init__(self, connection, header, debug):
        self.connection = connection
        self.header = header[0]
        self.debug = debug
    
    def read(self, byte: int) -> HanazederPacket:
        finished_packet = self.handle_byte(byte)
        if finished_packet:
            return finished_packet

    def handle_byte(self, byte: int) -> HanazederPacket:
        if self.state != ReaderState.LOOKING_FOR_HEADER \
                and self.state != ReaderState.EXPECTING_CHECKSUM \
                and self.state != ReaderState.ESCAPING:
            self.crc.process(byte.to_bytes(1, byteorder='little'))
        if self.debug:
            logger.debug('Handle_byte %X state in %s', byte, self.state.name)
        
        if byte == self.header and self.state != ReaderState.LOOKING_FOR_HEADER \
                and self.state != ReaderState.ESCAPING:
            if self.debug:
                logger.debug('Found escape byte, skipping')
            self.state_before_escaping = self.state
            self.state = ReaderState.ESCAPING
            return
        
        if self.state == ReaderState.ESCAPING:
            self.state = self.state_before_escaping
            self.state_before_escaping = None

        if self.state == ReaderState.LOOKING_FOR_HEADER:
            if byte == self.header:
                self.packet = HanazederPacket()
                self.state = ReaderState.EXPECTING_MSG_NO
                self.crc = Crc8Maxim()
        elif self.state == ReaderState.EXPECTING_MSG_NO:
            self.packet.msg_no = byte
            self.state = ReaderState.EXPECTING_TYPE
        elif self.state == ReaderState.EXPECTING_TYPE:
            self.packet.msg_type = byte
            self.state = ReaderState.EXPECTING_SIZE
        elif self.state == ReaderState.EXPECTING_SIZE:
            self.packet.msg_size = byte
            self.packet.msg = bytearray()
            if self.packet.msg_size == 0:
                self.state = ReaderState.EXPECTING_CHECKSUM
            else:
                self.state = ReaderState.READING_PAYLOAD
        elif self.state == ReaderState.READING_PAYLOAD or self.state == ReaderState.ESCAPING:
            self.packet.msg += bytearray(byte.to_bytes(1, byteorder='little'))
            self.packet.msg_size = self.packet.msg_size - 1
            if self.packet.msg_size <= 0:
                self.state = ReaderState.EXPECTING_CHECKSUM
            else:
                self.state = ReaderState.READING_PAYLOAD
        elif self.state == ReaderState.EXPECTING_CHECKSUM:
            self.state = ReaderState.LOOKING_FOR_HEADER
            calculated_crc = self.crc.finalbytes()
            if calculated_crc[0] != byte:
                logger.error('Wrong checksum')
            else:
                # Packet fully read
                return self.packet
        if self.debug:
            logger.debug('Handle_byte state out %s', self.state.name)


def hanazeder_decode_num(header, value) -> float:
    if value == SENSOR_GONE:
        return None
    int_val = int.from_bytes(value, byteorder='little', signed=True)
    return int_val / 10

def hanazeder_decode_byte(byte: bytes) -> int:
    return int.from_bytes(byte, byteorder='little')

def hanazeder_read(expected_header: bytes, msg_num: int, connection: SerialOrNetwork) -> bytes:
    # start by reading at least four bytes
    header = connection.read(4)
    if len(header) < 4:
        raise ReadtimeoutException("Could not read header")
    if header[0] is not expected_header[0]:
        raise InvalidHeaderException(f'Expecting header {expected_header[0]}, got {header[0]} instead')
    if header[1] != msg_num:
        raise InvalidHeaderException(f'Expecting message number {msg_num} but got {header[1]}')
    value_size = header[3]
    value = connection.read(value_size)
    # Always followed by one byte checksum
    checksum = connection.read(1)
    crc = Crc8Maxim()
    crc.process(header[1:])
    crc.process(value)
    calculated_crc = crc.finalbytes()
    if calculated_crc != checksum:
        raise ChecksumNotMatchingException(f'Expected checksum {calculated_crc} but got {checksum}')
    # TODO: unescape header in value
    return value

def hanazeder_encode_msg(header: bytes, msg_num: int, request: bytes) -> bytes:
    if msg_num < 0 or msg_num > 255:
        raise IllegalArgumentException("msg_num must be between 0 and 255")
    if len(header) != 1:
        raise IllegalArgumentException("header must be single byte")
    # Checksum is without header
    checksum_msg = bytearray()
    checksum_msg.append(msg_num)
    checksum_msg += bytearray(request)
    checksum = Crc8Maxim.calcbytes(checksum_msg)
    # Escape header in request
    escaped_msg = bytearray(header) + bytearray(checksum_msg.replace(b'\xEE', b'\xEE\xEE')) + bytearray(checksum)
    # msg += bytearray(checksum)
    
    return bytes(escaped_msg)


def hex_to_byte(hexStr: str) -> bytes:
    """
    Convert a string hex byte values into a byte string. The Hex Byte values may
    or may not be space separated.
    """
    hexStr = ''.join(hexStr.split(" "))
    return bytes.fromhex(hexStr)
	
def byte_to_hex(bytes: bytes) -> str:
    """
    Convert a byte string to it's hex string representation e.g. for output.
    """
    return bytes.hex()

def dec_to_bytes(decimal: int):
    """return the hexadecimal string representation of integer n"""
    return decimal.to_bytes(2, byteorder='little')